
import java.awt.*;
import java.lang.Math;
import java.util.*;
import MDDemo;
import MDInterface;
import Atom;

public class Simulation extends Thread {
    public static final int N_ATOMS = 100;  //must be a perfect square
    MDInterface mdinterface = null;
    public LinkedList atoms;
    private Random rand;
    public double temp;
    public double regionEdge;
    public double hregionEdge;
    public double pe;
    public double ke;
    public double te;
    public int ts = 0;
    private double mass = 39.948;

    // Configurable parameters (by user)
    private double density;
    private boolean scaling = true;
    private double applTemp;
    private double epsilon = 398.4;
    private double sigma = 3.41;
    private double cutoff = 8.5;

    static final double AVOGAD = 6.0221367e23;
    static final double MASSUNIT = 1.6605402e-27;
    static final double LENGTHUNIT = 1.0e-10;
    static final double TIMEUNIT = 1.0e-12;
    static final double ENERGYUNIT = (MASSUNIT * (LENGTHUNIT / TIMEUNIT) *
				    (LENGTHUNIT / TIMEUNIT));
    static final double KB = 1.380658e-23 / ENERGYUNIT;
    static final double TIMESTEP = 0.01;

    private double[] keArray = new double[10];
    private int kePointer = 0;

  public Simulation(double t, double d) {
    Atom a;
    double x,y,cellEdge;
    double vx,vy;
    int atomd;
    
    atomd = (int)Math.sqrt((double)N_ATOMS);
    this.density = d;
    this.applTemp = t;
    regionEdge = Math.sqrt((double)N_ATOMS/this.density);
    hregionEdge = 0.5 * regionEdge;
    cellEdge = regionEdge/atomd;
    rand = new Random();
    atoms = new LinkedList();
    for(int i = 0; i < atomd; i++) {
      for(int j = 0 ; j < atomd ; j++) {
	x = (((double)i + 0.5) * cellEdge) - hregionEdge;
	y = (((double)j + 0.5) * cellEdge) - hregionEdge;
	vx = Math.sqrt( (KB * applTemp) / mass ) * rand.nextGaussian();
	vy = Math.sqrt( (KB * applTemp) / mass ) * rand.nextGaussian();
	a = new Atom(x,y,vx,vy);
	atoms.append(a);
      }
    }
  }

  public void run() {
    calcForces();
    measurements();
    update();
    while(true) {
      integrate_step1();
      zero_forces();
      calcForces(); 
      integrate_step2();
      boundary_cond();
      measurements();
      if(ts % 10 == 0) {
	  if(scaling) scaleTemp();
      }
      ts++;
      update();
      updateDensity();
      try {sleep(20);} catch (InterruptedException e) {}
    } 
  }

  public void update() {
      mdinterface.update();
  }

  private void calcForces() {
    Enumeration e1,e2;
    Atom a1,a2;
    double rij2,rij;
    double rijx,rijy;
    double ri2,ri12,ri6;
    double cutoff2 = cutoff * cutoff;
    double force;
    
    e1 = atoms.elements();
    e2 = atoms.elements();
    pe = 0.0;

    while(e1.hasMoreElements()) {
      a1 = (Atom)e1.nextElement();
      e2 = atoms.elements();
      while(e2.hasMoreElements()) {
        a2 = (Atom)e2.nextElement();
        if(a2 != a1) {
          rijx = a1.x - a2.x;
          if(rijx > hregionEdge) rijx = rijx - regionEdge;
          if(rijx < -hregionEdge) rijx = rijx + regionEdge;
          rijy = a1.y - a2.y;
          if(rijy > hregionEdge) rijy = rijy - regionEdge;
          if(rijy < -hregionEdge) rijy = rijy + regionEdge;
          rij2 = (rijx * rijx) + (rijy * rijy);
          if(rij2 < cutoff2) {
            ri2 = 1.0 / rij2;
            ri6 = sigma * sigma * ri2;
            ri6 = ri6 * ri6 * ri6;
            ri12 = ri6 * ri6;
            pe += epsilon * (ri12 - ri6);
            force = ri2 * 6.0 * epsilon * ((2.0 * ri12) - ri6);
            a1.fx += force * rijx;
            a1.fy += force * rijy;
          }
        }
      }
    }
    pe = 0.5 * pe;
  }

  private void zero_forces() {
    Enumeration e;
    Atom a;

    e = atoms.elements();
    while(e.hasMoreElements()) {
      a = (Atom)e.nextElement();
      a.fx = 0.0;
      a.fy = 0.0;
    }
  }

  private void integrate_step1() {
    Enumeration e;
    Atom a;
    double dt2 = TIMESTEP * TIMESTEP;

    e = atoms.elements();
    while(e.hasMoreElements()) {
      a = (Atom)e.nextElement();
      a.x = a.x + (TIMESTEP * a.vx) + (0.5 * dt2 * (a.fx / mass));
      a.y = a.y + (TIMESTEP * a.vy) + (0.5 * dt2 * (a.fy / mass));
      a.vx = a.vx + (0.5 * TIMESTEP * (a.fx / mass));
      a.vy = a.vy + (0.5 * TIMESTEP * (a.fy / mass));
    }
  }

  private void integrate_step2() {
    Enumeration e;
    Atom a;

    e = atoms.elements();
    while(e.hasMoreElements()) {
      a = (Atom)e.nextElement();
      a.vx = a.vx + (0.5 * TIMESTEP * (a.fx / mass));
      a.vy = a.vy + (0.5 * TIMESTEP * (a.fy / mass));
    }
  }

  private void boundary_cond() {
    Enumeration e;
    Atom a;

    e = atoms.elements();
    while(e.hasMoreElements()) {
      a = (Atom)e.nextElement();
      if(a.x > hregionEdge) a.x = a.x - regionEdge;
      if(a.x <= -hregionEdge) a.x = a.x + regionEdge;
      if(a.y > hregionEdge) a.y = a.y - regionEdge;
      if(a.y <= -hregionEdge) a.y = a.y + regionEdge;
    }
  }

  private void measurements() {
    Enumeration e;
    Atom a;
    double vsqr = 0.0;

    ke = 0.0;
    te = 0.0;
    e = atoms.elements();
    while(e.hasMoreElements()) {
      a = (Atom)e.nextElement();
      vsqr += (a.vx * a.vx) + (a.vy * a.vy);
    }
    ke = 0.5 * mass * vsqr;
    te = ke + pe;
    temp = ke / (N_ATOMS * KB);
    keArray[kePointer] = ke;
    kePointer++;
    kePointer = kePointer % 10;
  }

    private void scaleTemp() {
	double sum = 0.0,scaleFactor;
	Enumeration e;
	Atom a;

	if(ts > 10) {
	    for (int i = 0; i<10; i++){
		sum += keArray[i];
	    }
	    sum = sum / 10;
	    scaleFactor = Math.sqrt((KB * N_ATOMS * applTemp) / sum);
	    e = atoms.elements();
	    while(e.hasMoreElements()) {
		a = (Atom)e.nextElement();
		a.vx = a.vx * scaleFactor;
		a.vy = a.vy * scaleFactor;
	    }
	}
    }

  private void updateDensity() {
    Enumeration e;
    Atom a;
    double factor;
    double oldedge,new_dens;

    new_dens = this.mdinterface.getDensity();
    if (new_dens != density) {
	oldedge = regionEdge;
	e = atoms.elements();
	regionEdge = Math.sqrt((double)N_ATOMS / new_dens);
	hregionEdge = 0.5 * regionEdge;
	density = new_dens;
	
	factor = regionEdge / oldedge;
	while(e.hasMoreElements()) {
	    a = (Atom)e.nextElement();
	    a.x = a.x * factor;
	    a.y = a.y * factor;
	}
	mdinterface.setEdgeLabel(regionEdge);
    }
  }

  public void setInterface(MDInterface p) {
    this.mdinterface = p;
  }

  public void resetSim() {
    Enumeration e;
    Atom a;
    double x,y,vx,vy;
    int atomd = (int)Math.sqrt((double)N_ATOMS);
    double cellEdge = regionEdge/atomd;

    e = atoms.elements();

    for(int i = 0; i < atomd; i++) {
      for(int j = 0 ; j < atomd ; j++) {
	x = (((double)i + 0.5) * cellEdge) - hregionEdge;
	y = (((double)j + 0.5) * cellEdge) - hregionEdge;
	vx = Math.sqrt( (KB * applTemp) / mass ) * rand.nextGaussian();
	vy = Math.sqrt( (KB * applTemp) / mass ) * rand.nextGaussian();
	a = (Atom)e.nextElement();
	a.x = x;
	a.y = y;
	a.vx = vx;
	a.vy = vy;
	a.fx = 0.0;
	a.fy = 0.0;
      }
    } 
    mdinterface.reset();
  }

    public double getEpsilon(){
	return epsilon;
    }

    public double getSigma() {
	return sigma;
    }

    public double getCutoff() {
	return cutoff;
    }

    public synchronized void setCutoff(double c) {
	this.cutoff = c;
    }

    public synchronized void setSigma(double n) {
	this.sigma = n;
    }

    public synchronized void setEpsilon(double n) {
	this.epsilon = n;
    }
  
    public synchronized void setScaling(boolean state) {
	scaling = state;
    }

    public synchronized void setApplTemp(double temp) {
	applTemp = temp;
    }


}

